/*
 * m_gline.c: Server-only, persistent (sent in burst) GLINE support
 *
 * Copyright (c) 2004  Ben Winslow.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    the above paragraph (the one permitting redistribution), this list of
 *    conditions and the following disclaimer in the documentation and/or
 *    other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id$
 */

/*
 * Server GLINE protocol:
 *
 * Add:				GLINE +user@host TS LENGTH :reason
 * Remove:			GLINE -user@host TS :reason
 * Netburst:		GLINE @ TS
 * Netburst end:	GLINE $ TS
 */

#include "stdinc.h"
#include "msg.h"
#include "handlers.h"
#include "s_log.h"
#include "client.h"
#include "s_conf.h"
#include "s_serv.h"
#include "s_gline.h"
#include "common.h"
#include "send.h"
#include "parse.h"
#include "list.h"
#include "hook.h"
#include "modules.h"

static void ms_gline(struct Client *, struct Client *, int, char *[]);
static void gline_add(struct Client *, char *, time_t, time_t, const char *);
static void gline_remove(struct Client *, char *, time_t, const char *);
static void gline_got_ts(struct Client *, time_t, int);
#ifdef STATIC_MODULES
int gline_netburst(struct Client *);
#else
static int gline_netburst(struct Client *);
#endif
static void gline_list_clear(void);
static void gline_list_send(struct Client *);
static int gline_find(const char *, const char *);

struct Message gline_msgtab = {
	"GLINE", 0, 0, 3, 5, 0, 0,
		{ m_unregistered, m_ignore, ms_gline, m_ignore }
};

#ifndef STATIC_MODULES
void _modinit(void)
{
	mod_add_cmd(&gline_msgtab);
	hook_add_hook("burst", (hookfn *)gline_netburst);
}

void _moddeinit(void)
{
	hook_del_hook("burst", (hookfn *)gline_netburst);
	mod_del_cmd(&gline_msgtab);
}

const char *_version = "$Revision: 1.0 $";
#endif

/*
 * ms_gline: handle a GLINE command from a server
 *
 * notes: parv[1] may be modified
 *
 * parv[0] - originating server
 * parv[1] - [+-]user@host, @, or $
 * parv[2] - TS
 * parv[3] - length (+), reason (-)
 * parv[4] - reason (+)
 *
 * GLINEs from servers that are neither hubs nor listed in a U: line will
 * be dropped (and a warning issued)
 */
void ms_gline(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	time_t ts, len;

	if (!IsCapable(client_p, CAP_HUB) &&
		!find_u_conf(client_p->name, NULL, NULL)) {
		
		ilog(L_WARN, "Received GLINE from non-hub/non-shared server %s: %s",
			client_p->name, parv[1]);
		sendto_realops_flags(FLAGS_ALL, L_ALL, "Received G-LINE from non-hub/non-shared server %s: %s",
			client_p->name, parv[1]);

		return;
	}

	ts = atol(parv[2]);
	
	switch (*parv[1]) {
		case '+':
			if (parc < 5)
				return;

			len = atol(parv[3]);
			gline_add(client_p, parv[1] + 1, ts, len, parv[4]);

			break;

		case '-':
			if (parc < 4)
				return;

			gline_remove(client_p, parv[1] + 1, ts, parv[3]);

			break;

		case '@':
		case '$':
			gline_got_ts(client_p, ts, *parv[1] == '$');
			break;

		default: /* just drop GLINE commands we don't know how to handle */
			return;
	}
}

static void gline_add(struct Client *client_p, char *userhost, time_t ts, time_t len, const char *reason)
{
	struct ConfItem *gline;
	char *user, *host;

	if ((host = strchr(userhost, '@'))) {
		user = userhost;

		*host = 0;
		host++;
	} else
		return;

	ServerInfo.gline_ts = ts;
	
	/* if it's already glined, we're probably merging GLINE lists */
	if (gline_find(user, host))
		return;

	/* it's new, propogate it */
	sendto_server(client_p, NULL, NULL, CAP_GLN, NOCAPS, NOFLAGS,
		"GLINE +%s@%s %ld %ld :%s", user, host, ts, len, reason);

	gline = make_conf();
	gline->status = CONF_KILL;
	if (len == 0)
		gline->hold = 0;					/* never expire */
	else
		gline->hold = CurrentTime + len;	/* expire time */
	gline->ts = ts;							/* timestamp */
	DupString(gline->passwd, reason);		/* reason */
	DupString(gline->name, user);			/* user */
	DupString(gline->host, host);			/* host */

	add_gline(gline);
	
	check_klines();
}

static void gline_remove(struct Client *client_p, char *userhost, time_t ts, const char *reason)
{
	char *user, *host;

	if ((host = strchr(userhost, '@'))) {
		user = userhost;

		*host = 0;
		host++;
	} else
		return;

	ServerInfo.gline_ts = ts;

	/* we always propogate - GLINEs, just in case */
	sendto_server(client_p, NULL, NULL, CAP_GLN, NOCAPS, NOFLAGS,
		"GLINE -%s@%s %ld :%s", user, host, ts, reason);

	remove_gline_match(user, host);
}

/*
 * gline_got_ts: handle receipt of GLINE timestamp
 *
 * brief description of sementics:
 * We're a hub and foreign is a leaf: Always send our GLINE list
 * We're a leaf and foreign is a hub: Always clear our GLINE list and wait
 *                                    for the hub to send theirs.
 * We're equal to the foreign server: Compare timestamps:
 *     Our timestamp is newer: Send our GLINE list
 *     Our timestamp is older: Clear our GLINE list, wait for theirs
 *     Our timestamp is equal: Send our GLINE list (lists will be merged)
 */
static void gline_got_ts(struct Client *client_p, time_t ts, int final)
{
	/* end-of-burst GLINE timestamp - just update the TS */
	if (final) {
		sendto_server(client_p, NULL, NULL, CAP_GLN, NOCAPS, NOFLAGS,
			"GLINE %c %ld", '$', ts);
		ServerInfo.gline_ts = ts;
		return;
	}

	if (ServerInfo.hub && !IsCapable(client_p, CAP_HUB)) {
		gline_list_send(client_p);
	} else if (!ServerInfo.hub && IsCapable(client_p, CAP_HUB)) {
		gline_list_clear();
		ServerInfo.gline_ts = ts;
	} else {
		if (ts > ServerInfo.gline_ts) {
			/* tell our leaves to wipe their GLINEs, too */
			sendto_server(client_p, NULL, NULL, CAP_GLN, NOCAPS, NOFLAGS,
				"GLINE %c %ld", '@', ts);
			gline_list_clear();
			ServerInfo.gline_ts = ts;
		} else if (ts <= ServerInfo.gline_ts) {
			gline_list_send(client_p);
		}
	}
}

#ifdef STATIC_MODULES
int gline_netburst(struct Client *client_p)
#else
static int gline_netburst(struct Client *client_p)
#endif
{
	if (!IsCapable(client_p, CAP_GLN))
		return 0;

	/* We don't need to even bother sending a GLINE command if we're a leaf */
	if (ServerInfo.hub) {
		/* Leaves can't define GLINEs, so we shove ours to them */
		if (!IsCapable(client_p, CAP_HUB))
			gline_list_send(client_p);
		else
			sendto_one(client_p, "GLINE @ %ld", ServerInfo.gline_ts);
	} else
		gline_list_clear();

	return 0;
}

static void gline_list_clear(void)
{
	dlink_node *cur, *next;
	
	DLINK_FOREACH_SAFE(cur, next, glines.head) {
		free_conf(cur->data);
		dlinkDelete(cur, &glines);
		free_dlink_node(cur);
	}
}

static void gline_list_send(struct Client *client_p)
{
	dlink_node *cur;
	struct ConfItem *gline;
	time_t remaining;

	DLINK_FOREACH(cur, glines.head) {
		gline = cur->data;

		if (gline->hold == 0)
			remaining = 0;
		else if (gline->hold - CurrentTime <= 0)
			continue;
		else
			remaining = gline->hold - CurrentTime;

		sendto_one(client_p, "GLINE +%s@%s %ld %ld :%s",
			gline->name, gline->host, gline->ts, remaining, gline->passwd);
	}
	/*
	 * send the TS in case we've removed a GLINE with a newer TS than we
	 * sent -- the other end will reset the TS whenever they get a GLINE
	 * command.  Redundant timestamp on $ to make parsing simpler.
	 */
	
	sendto_one(client_p, "GLINE $ %ld", ServerInfo.gline_ts);
}

static int gline_find(const char *user, const char *host)
{
	dlink_node *cur;
	struct ConfItem *gline;
	
	DLINK_FOREACH(cur, glines.head) {
		gline = cur->data;
		
		if (!irccmp(user, gline->name) && !irccmp(host, gline->host))
			return 1;
	}
	return 0;
}
