/*
 * isupport.c: Extended ISUPPORT numeric support
 *
 * Copyright (c) 2004  Ben Winslow.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    the above paragraph (the one permitting redistribution), this list of
 *    conditions and the following disclaimer in the documentation and/or
 *    other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id$
 */

#include "config.h"
#include "stdinc.h"
#include "event.h"
#include "ircd_defs.h"
#include "numeric.h"
#include "client.h"
#include "ircd.h"
#include "supported.h" /* for CASEMAP */
#include "s_conf.h"
#include "channel_mode.h"
#include "s_log.h"
#include "hook.h"
#include "send.h"
#ifdef SSL_CLIENT_SUPPORT
#include "ssl_client.h"
#endif
#include "isupport.h"

static int isupport_add_token(char *, struct isupport_s *, int, int *);
static void isupport_build(int);
static int isupport_update(int);

static struct isupport_s isup;

void isupport_init(void)
{
	memset(&isup, 0, sizeof(isup));
	DupString(isup.old_netname, NETWORK_NAME_DEFAULT);

	hook_add_hook("update_isupport", (hookfn *)isupport_update);
}

void isupport_send(struct Client *client_p)
{
	int x;

	for (x = 0; x < isup.lines; x++) {
		sendto_one(client_p, form_str(RPL_ISUPPORT), me.name,
			client_p->name, isup.line[x]);
	}
}

void isupport_deinit(void)
{
	if (isup.old_netname != NULL)
		MyFree(isup.old_netname);
}

/* hook function */
static int isupport_update(int send_changed)
{
	isupport_build(send_changed);
	
	return 0;
}

static int isupport_add_token(char *buf, struct isupport_s *is, int maxlen, int *toknum)
{
	if (is == NULL || toknum == NULL || is->lines > MAX_ISUPPORT_LINES)
		return -1;

	/*
	 * if the token won't fit on this line, or we've hit the token
	 * limit, advance to the next isupport line
	 */
	
	if (strlen(is->line[is->lines]) + 1 + strlen(buf) > maxlen ||
		*toknum >= 13) {
		
		is->lines++;
		(*toknum) = 0;
		
		if (is->lines >= MAX_ISUPPORT_LINES) {
			ilog(L_ERROR, "ISUPPORT reply too large -- Increase MAX_ISUPPORT_LINES");
			return -1;
		}
	}
	
	/* add a space if this isn't the first token */
	if (is->line[is->lines][0])
		strcat(is->line[is->lines], " ");

	/* add the token */
	strcat(is->line[is->lines], buf);
	
	(*toknum)++;

	return 0;
}

static void isupport_build(int send_changed)
{
	int maxlen, tok = 0, tok_changed = 0, x;
	char curtoken[BUFSIZE];
	struct isupport_s changed;
	dlink_node *dlnode, *dlnode_next;
	struct Client *client_p;

	for (x = 0; x < MAX_ISUPPORT_LINES; x++)
		memset(isup.line[x], 0, sizeof(isup.line[x]));
	isup.lines = 0;

	memset(&changed, 0, sizeof(changed));
	
	sprintf(curtoken, form_str(RPL_ISUPPORT), me.name, "", "");
	maxlen = BUFSIZE - (strlen(curtoken) + NICKLEN + 3);
	
	/* rain - ISUPPORT tokens from draft-brocklesby-irc-isupport-03 */

	/* *** ISUPPORT TOKEN: STD */
	sprintf(curtoken, "STD=%s", ISUPPORT_STANDARD);
	isupport_add_token(curtoken, &isup, maxlen, &tok);

	/* *** ISUPPORT TOKEN: CASEMAPPING */
	sprintf(curtoken, "CASEMAPPING=%s", CASEMAP);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	
	/* *** ISUPPORT TOKEN: CHANLIMIT */
	sprintf(curtoken, "CHANLIMIT=%s:%d",
		ConfigServerHide.disable_local_channels ? "#" : "#&",
		ConfigChannel.max_chans_per_user);
	isupport_add_token(curtoken, &isup, maxlen, &tok);

	if (ConfigServerHide.disable_local_channels !=
		((isup.old_flags & ISUP_FLAG_NOLOCAL) ? 1 : 0) ||
		ConfigChannel.max_chans_per_user != isup.old_maxchans) {
		
		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		
		/* ISUP_FLAG_NOLOCAL set in CHANTYPES */
		/* old_maxchans set in MAXCHANNELS */
	}
	
	/* *** ISUPPORT TOKEN: CHANMODES */
	sprintf(curtoken, "CHANMODES=b%s%s,k,l,imnpst%s",
		ConfigChannel.use_except ? "e" : "",
		ConfigChannel.use_invex ? "I" : "",
		ConfigChannel.use_anonops ? "a" : "");
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	
	if (ConfigChannel.use_except != !!(isup.old_flags & ISUP_FLAG_EXCEPTS) ||
		(ConfigChannel.use_invex != !!(isup.old_flags & ISUP_FLAG_INVEX))) {
		
		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		
		/* ISUP_FLAG_EXCEPTS set in EXCEPT */
		/* ISUP_FLAG_INVEX set in INVEX */
	}
	
	/* *** ISUPPORT TOKEN: MAXLIST */
	sprintf(curtoken, "MAXLIST=b%s%s:%d",
		ConfigChannel.use_except ? "e" : "",
		ConfigChannel.use_invex ? "I" : "",
		ConfigChannel.max_bans);
	isupport_add_token(curtoken, &isup, maxlen, &tok);

	if (ConfigChannel.use_except != !!(isup.old_flags & ISUP_FLAG_EXCEPTS) ||
		(ConfigChannel.use_invex != !!(isup.old_flags & ISUP_FLAG_INVEX)) ||
		ConfigChannel.max_bans != isup.old_maxbans) {

		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		/* ISUP_FLAG_EXCEPTS set in EXCEPT */
		/* ISUP_FLAG_INVEX set in INVEX */
		/* old_maxbans is set in MAXBANS */
	}

	/* *** ISUPPORT TOKEN: EXCEPTS */
	if (ConfigChannel.use_except) {
		isupport_add_token("EXCEPTS", &isup, maxlen, &tok);
		if (!(isup.old_flags & ISUP_FLAG_EXCEPTS))
			isupport_add_token("EXCEPTS", &changed, maxlen, &tok_changed);
		isup.old_flags |= ISUP_FLAG_EXCEPTS;
	} else {
		if ((isup.old_flags & ISUP_FLAG_EXCEPTS))
			isupport_add_token("-EXCEPTS", &changed, maxlen, &tok_changed);
		isup.old_flags &= ~ISUP_FLAG_EXCEPTS;
	}

	/* *** ISUPPORT TOKEN: INVEX */
	if (ConfigChannel.use_invex) {
		isupport_add_token("INVEX", &isup, maxlen, &tok);
		if (!(isup.old_flags & ISUP_FLAG_INVEX))
			isupport_add_token("INVEX", &changed, maxlen, &tok_changed);
		isup.old_flags |= ISUP_FLAG_INVEX;
	} else {
		if ((isup.old_flags & ISUP_FLAG_INVEX))
			isupport_add_token("-INVEX", &changed, maxlen, &tok_changed);
		isup.old_flags &= ~ISUP_FLAG_INVEX;
	}
	
	/* *** ISUPPORT TOKEN: KICKLEN */
	sprintf(curtoken, "KICKLEN=%d", TOPICLEN);
	isupport_add_token(curtoken, &isup, maxlen, &tok);

	/* *** ISUPPORT TOKEN: MODES */
	sprintf(curtoken, "MODES=%d", MAXMODEPARAMS);
	isupport_add_token(curtoken, &isup, maxlen, &tok);

	/* *** ISUPPORT TOKEN: NETWORK */
	sprintf(curtoken, "NETWORK=%s", ServerInfo.network_name);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	if (strcmp(ServerInfo.network_name, isup.old_netname)) {
		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		MyFree(isup.old_netname);
		DupString(isup.old_netname, ServerInfo.network_name);
	}

	/* *** ISUPPORT TOKEN: NICKLEN */
	sprintf(curtoken, "NICKLEN=%d", NICKLEN - 1);
	isupport_add_token(curtoken, &isup, maxlen, &tok);

	/* *** ISUPPORT TOKEN: PREFIX */
	sprintf(curtoken, "PREFIX=%s",
		ConfigChannel.use_halfops ? "(ohv)@%+" : "(ov)@+");
	isupport_add_token(curtoken, &isup, maxlen, &tok);

	/* *** ISUPPORT TOKEN: STATUSMSG */
	sprintf(curtoken, "STATUSMSG=@%s+",
		ConfigChannel.use_halfops ? "%" : "");
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	
	/* *** ISUPPORT TOKEN: TARGMAX */
	sprintf(curtoken, "TARGMAX=PRIVMSG:%d,NOTICE:%d,ACCEPT:,JOIN:,PART:",
		ConfigFileEntry.max_targets, ConfigFileEntry.max_targets);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	if (ConfigFileEntry.max_targets != isup.old_maxtargets) {
		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		isup.old_maxtargets = ConfigFileEntry.max_targets;
	}

	/* *** ISUPPORT TOKEN: TOPICLEN */	
	sprintf(curtoken, "TOPICLEN=%d", TOPICLEN);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	
#ifdef SSL_CLIENT_SUPPORT
	/* *** ISUPPORT TOKEN: STARTTLS */
	if ((ServerInfo.ssl_status & SSLC_STAT_STARTTLS_OK) == SSLC_STAT_STARTTLS_OK) {
		isupport_add_token("STARTTLS", &isup, maxlen, &tok);
		if (!(isup.old_flags & ISUP_FLAG_STARTTLS))
			isupport_add_token("STARTTLS", &changed, maxlen, &tok_changed);
		isup.old_flags |= ISUP_FLAG_STARTTLS;
	} else {
		if ((isup.old_flags & ISUP_FLAG_STARTTLS))
			isupport_add_token("-STARTTLS", &changed, maxlen, &tok_changed);
		isup.old_flags &= ~ISUP_FLAG_STARTTLS;
	}
#endif

#if CHANNELLEN != 200
	/* *** ISUPPORT TOKEN: CHANNELLEN */
	sprintf(curtoken, "CHANNELLEN=%d", CHANNELLEN);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
#endif

	/* *** ISUPPORT TOKEN: CHANTYPES */
	/* rain - don't send CHANTYPES if it's default */
	if (ConfigServerHide.disable_local_channels) {
		isupport_add_token("CHANTYPES=#", &isup, maxlen, &tok);
		if (!(isup.old_flags & ISUP_FLAG_NOLOCAL))
			isupport_add_token("CHANTYPES=#", &changed, maxlen, &tok_changed);
		isup.old_flags |= ISUP_FLAG_NOLOCAL;
	} else {
		if ((isup.old_flags & ISUP_FLAG_NOLOCAL))
			isupport_add_token("-CHANTYPES", &changed, maxlen, &tok_changed);
		isup.old_flags &= ~ISUP_FLAG_NOLOCAL;
	}

	/* *** ISUPPORT TOKEN: KNOCK */
	/* rain - misc. stuff that isn't defined in the draft */
	if (ConfigChannel.use_knock) {
		isupport_add_token("KNOCK", &isup, maxlen, &tok);
		if (!(isup.old_flags & ISUP_FLAG_KNOCK))
			isupport_add_token("KNOCK", &changed, maxlen, &tok_changed);
		isup.old_flags |= ISUP_FLAG_KNOCK;
	} else {
		if ((isup.old_flags & ISUP_FLAG_KNOCK))
			isupport_add_token("-KNOCK", &changed, maxlen, &tok_changed);
		isup.old_flags &= ~ISUP_FLAG_KNOCK;
	}
	
	/* *** ISUPPORT TOKEN: VCHANS */
	if (ConfigChannel.use_vchans) {
		isupport_add_token("VCHANS", &isup, maxlen, &tok);
		if (!(isup.old_flags & ISUP_FLAG_VCHANS))
			isupport_add_token("VCHANS", &changed, maxlen, &tok_changed);
		isup.old_flags |= ISUP_FLAG_VCHANS;
	} else {
		if ((isup.old_flags & ISUP_FLAG_VCHANS))
			isupport_add_token("-VCHANS", &changed, maxlen, &tok_changed);
		isup.old_flags &= ~ISUP_FLAG_VCHANS;
	}

	/* *** ISUPPORT TOKEN: CALLERID */
	isupport_add_token("CALLERID=g", &isup, maxlen, &tok);


	/* rain - these are sent for reverse compatibility (for now) */

	/* *** ISUPPORT TOKEN: MAXCHANNELS */
	sprintf(curtoken, "MAXCHANNELS=%d", ConfigChannel.max_chans_per_user);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	if (ConfigChannel.max_chans_per_user != isup.old_maxchans) {
		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		isup.old_maxchans = ConfigChannel.max_chans_per_user;
	}
	
	/* *** ISUPPORT TOKEN: MAXBANS */
	sprintf(curtoken, "MAXBANS=%d", ConfigChannel.max_bans);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	if (ConfigChannel.max_bans != isup.old_maxbans) {
		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		isup.old_maxbans = ConfigChannel.max_bans;
	}
	
	/* *** ISUPPORT TOKEN: WALLCHOPS */
	isupport_add_token("WALLCHOPS", &isup, maxlen, &tok);
	
	/* *** ISUPPORT TOKEN: MAXTARGETS */
	sprintf(curtoken, "MAXTARGETS=%d", ConfigFileEntry.max_targets);
	isupport_add_token(curtoken, &isup, maxlen, &tok);
	if (ConfigFileEntry.max_targets != isup.old_maxtargets) {
		isupport_add_token(curtoken, &changed, maxlen, &tok_changed);
		isup.old_maxtargets = ConfigFileEntry.max_targets;
	}

	/* isup must always have at least one line */
	isup.lines += 1;
	/* changed might not, so check */
	if (changed.line[0][0])
		changed.lines += 1;

	if (send_changed) {
		DLINK_FOREACH_SAFE(dlnode, dlnode_next, lclient_list.head) {
			client_p = dlnode->data;
			for (x = 0; x < changed.lines; x++) {
				sendto_one(client_p, form_str(RPL_ISUPPORT),
					me.name, client_p->name, changed.line[x]);
			}
		}
	}
}
