/*
 * ssl_client.c: SSL communications support for ircd-hybrid
 *
 * Copyright (c) 2004  Ben Winslow.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    the above paragraph (the one permitting redistribution), this list of
 *    conditions and the following disclaimer in the documentation and/or
 *    other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id$
 */

#include "config.h"

#ifdef SSL_CLIENT_SUPPORT
#include "stdinc.h"
#include "ssl_client.h"
#include "client.h"
#include "s_log.h"
#include "s_conf.h"
#include "s_user.h"
#include "send.h"

static void SSLC_VerifyKeys(void);

static SSL_CTX *ssl_ctx;
static const unsigned char *ssl_sid_context = "ircd"; /* arbitrary */

/* TODO: add a password callback so that the PEM can be encrypted? */

/* Init the OpenSSL context */
void SSLC_CTX_init(void)
{
	SSL_load_error_strings();
	SSL_library_init();
	
	if ((ssl_ctx = SSL_CTX_new(SSLv23_server_method())) == NULL) {
		ERR_print_errors_fp(stderr);
		exit(EXIT_FAILURE);
	}
	
	if (SSL_CTX_set_session_id_context(ssl_ctx, ssl_sid_context,
		strlen(ssl_sid_context)) != 1) {
		
		ERR_print_errors_fp(stderr);
		exit(EXIT_FAILURE);
	}
	
	SSL_CTX_set_session_cache_mode(ssl_ctx, SSL_SESS_CACHE_SERVER);

	/* 300s default - increase this or make it configurable? */
/*	SSL_CTX_set_timeout(ssl_ctx, 300); */
}

/* Deinitialize SSL context */
void SSLC_CTX_deinit(void)
{
	SSL_CTX_free(ssl_ctx);
}

/*
 * Perform the initial SSL handshake with a connecting client, allocating
 * the SSL data structures if necessary.
 */
int SSLC_handshake(struct Client *client_p)
{
	struct LocalUser *lclient_p;
	unsigned long ssl_err, ssl_errno;

	if (client_p == NULL)
		return SSLC_HANDSHAKE_INVALID_PARAM;

	if ((lclient_p = client_p->localClient) == NULL)
		return SSLC_HANDSHAKE_INVALID_PARAM;

	/* set up client data */
	if (lclient_p->ssl == NULL) {
		if ((lclient_p->ssl = SSL_new(ssl_ctx)) == NULL) {
			ilog(L_WARN, "SSL_new() failure for %s: %s",
				get_client_name(client_p, SHOW_IP),
				ERR_error_string(ERR_get_error(), NULL));
			return SSLC_HANDSHAKE_NEW_FAIL;
		}

		if (SSL_set_fd(lclient_p->ssl, lclient_p->fd) != 1) {
			ilog(L_WARN, "SSL_set_fd() failure for %s: %s",
				get_client_name(client_p, SHOW_IP),
				ERR_error_string(ERR_get_error(), NULL));
			return SSLC_HANDSHAKE_SETFD_FAIL;
		}
	}

	/* do SSL negotiation */
	if ((ssl_err = SSL_accept(lclient_p->ssl)) != 1) {
		ssl_errno = SSL_get_error(lclient_p->ssl, ssl_err);

		if (ssl_errno != SSL_ERROR_NONE) {
			if (ssl_errno == SSL_ERROR_WANT_READ || ssl_errno == SSL_ERROR_WANT_WRITE)
				return SSLC_HANDSHAKE_NOERROR;

			/*
			 * reset the SSL state so that the client gets a plaintext
			 * error message when they get dropped
			 */
			lclient_p->ssl_state = SSLC_STATE_NONE;

			if (ssl_errno == SSL_ERROR_SSL) {
				ilog(L_NOTICE, "SSL_accept() for %s failed: %s",
					get_client_name(client_p, SHOW_IP),
					ERR_error_string(ERR_get_error(), NULL));
					
				return SSLC_HANDSHAKE_SSLERROR;
			} else if (ssl_errno == SSL_ERROR_SYSCALL) {
				ilog(L_NOTICE, "SSL_accept() for %s failed: %s",
					get_client_name(client_p, SHOW_IP),
					strerror(errno));
			} else {
				ilog(L_NOTICE, "SSL_accept() for %s failed: SSL error %ld",
					get_client_name(client_p, SHOW_IP),
					ssl_errno);
			}

			return SSLC_HANDSHAKE_ACCEPT_FAIL;
		}
	}

	/* if they're already registered, add the mode */
	if (IsRegistered(client_p)) {
		int old = client_p->umodes & ALL_UMODES;
		
		client_p->umodes |= FLAGS_SSL;
		send_umode_out(client_p, client_p, old);
	}
	lclient_p->ssl_state = SSLC_STATE_CONNECTED;

	return 0;
}

/* Write bytes to an SSL client, or map failures to be POSIX-ish on failure */
int SSLC_write(struct Client *client_p, const void *buf, int num)
{
	struct LocalUser *lclient_p;
	int ret;

	if ((lclient_p = client_p->localClient) == NULL)
		return 0; /* major problems.  fake EOF. */
	
	if ((ret = SSL_write(lclient_p->ssl, buf, num)) < 0) {
		switch (SSL_get_error(lclient_p->ssl, ret)) {
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
				/*
				 * fake EWOULDBLOCK so that the socket code handles
				 * these cases properly
				 */
				errno = EWOULDBLOCK;
			default:
				break;
		}
	}
	
	return ret;
}

/* Read bytes from an SSL client, or map failures to be POSIX-ish on failure */
int SSLC_read(struct Client *client_p, void *buf, int num)
{
	struct LocalUser *lclient_p;
	int ret;

	if ((lclient_p = client_p->localClient) == NULL)
		return 0; /* major problems.  fake EOF. */
	
	if ((ret = SSL_read(lclient_p->ssl, buf, num)) < 0) {
		switch (SSL_get_error(lclient_p->ssl, ret)) {
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
				/*
				 * fake EWOULDBLOCK so that the socket code handles
				 * these cases properly
				 */
				errno = EWOULDBLOCK;
			default:
				break;
		}
	}
	
	return ret;
}

/* Shut down an SSL connection and free the related data structures */
void SSLC_shutdown(struct Client *client_p)
{
	struct LocalUser *lclient_p = client_p->localClient;

	if (lclient_p->ssl) {
		/* send our close notify -- don't bother waiting for the other end's */
		SSL_shutdown(lclient_p->ssl);
		SSL_free(lclient_p->ssl);
		
		lclient_p->ssl = NULL;
		lclient_p->ssl_state = SSLC_STATE_DEAD;
	}
}

/* Return a textual description of the cipher being used for a client */
const char *SSLC_cipherinfo(struct Client *client_p, int shortinfo)
{
	static char buf[512];
	SSL_CIPHER *cipher;
	int bits, processed_bits;

	cipher = SSL_get_current_cipher(client_p->localClient->ssl);

	if (shortinfo == 1) {
		bits = SSL_CIPHER_get_bits(cipher, &processed_bits);
		sprintf(buf, "%s %s (%d/%d bits)", SSL_CIPHER_get_version(cipher),
			SSL_CIPHER_get_name(cipher), processed_bits, bits);
	} else
		SSL_CIPHER_description(cipher, buf, sizeof(buf));

	return buf;
}

/* Attempt to load a certificate file */
void SSLC_CertificateFile(const char *fn)
{
	const char *err;
	
	if (SSL_CTX_use_certificate_chain_file(ssl_ctx, fn) != 1) {
		err = ERR_error_string(ERR_get_error(), NULL);

	    sendto_realops_flags(FLAGS_ALL, L_ALL,
	    	"Failed to load certificate '%s': %s", fn, err);
		ilog(L_ERROR, "Failed to load certificate '%s': %s", fn, err);
		
		ServerInfo.ssl_status &= ~SSLC_STAT_KEY_PUBLIC;
	} else
		ServerInfo.ssl_status |= SSLC_STAT_KEY_PUBLIC;

	SSLC_VerifyKeys();
}

/* Attempt to load a private key file */
void SSLC_PrivateKeyFile(const char *fn)
{
	const char *err;

	if (SSL_CTX_use_PrivateKey_file(ssl_ctx, fn, SSL_FILETYPE_PEM) != 1) {
		err = ERR_error_string(ERR_get_error(), NULL);
	    sendto_realops_flags(FLAGS_ALL, L_ALL,
	    	"Failed to load private key '%s': %s", fn, err);
		ilog(L_ERROR, "Failed to load private key '%s': %s", fn, err);
		
		ServerInfo.ssl_status &= ~SSLC_STAT_KEY_PRIVATE;
	} else
		ServerInfo.ssl_status |= SSLC_STAT_KEY_PRIVATE;

	SSLC_VerifyKeys();
}

/* Verify the certificate and private key, if both are loaded */
static void SSLC_VerifyKeys(void)
{
	if ((ServerInfo.ssl_status & SSLC_STAT_KEY_BOTH) == SSLC_STAT_KEY_BOTH) {
		if (SSL_CTX_check_private_key(ssl_ctx) != 1) {
			sendto_realops_flags(FLAGS_ALL, L_ALL,
				"Certificate and private key do not match");
			ilog(L_ERROR, "Certificate and private key do not match");
		}
	}
}

#else
#error not compiling with SSL support
#endif
